#! coding: utf-8
import re
import string

from lxml.html import diff

from .sentence import Sentence
from .placeholder import PlaceHolder

SENTENCES_END = u'.?!'
SPACE_REGEXP = re.compile('^(\s|&nbsp;)+$', flags=re.I | re.M)


def is_token_end(token):
    if SPACE_REGEXP.match(token.trailing_whitespace) or \
       token.trailing_whitespace in (u' ', u'\xa0') or \
       (token and token[-1] in string.punctuation) or \
       (token.post_tags and token.post_tags[-1][-1] == u' '):
        return True

    return False


def tokens_generator(tokens):
    result = []
    length = len(tokens) - 1

    for i, token in enumerate(tokens):
        result.append(token)
        next_one = None
        if i < length:
            next_one = tokens[i+1]
        if is_token_end(token) or (next_one and '<br>' in next_one.pre_tags) or (next_one and '<nbsp>' in next_one.pre_tags):
            yield result
            result = []
    else:
        yield result


def split_token(word_tokens):
    """
    Разбивает токены на куски например из ["привет.Как "] сделает [["привет."], ["Как"]]
    """
    results = []
    tokens = []
    for token in word_tokens:
        i = 0
        length = len(token)
        while i < length:
            if token[i] in SENTENCES_END:
                new_token = diff.token(token[:i+1])
                new_token.pre_tags = token.pre_tags
                if i == length - 1:
                    new_token.post_tags = token.post_tags
                    new_token.trailing_whitespace = token.trailing_whitespace
                tokens.append(new_token)
                results.append(tokens)
                tokens = []
                token = diff.token(token[i+1:], post_tags=token.post_tags,
                                   trailing_whitespace=token.trailing_whitespace)
                length = len(token)
                i = 0
            i += 1
        # если встретили пустой тег - скипаем его
        if len(token) == 0 and not token.pre_tags and not token.post_tags:
            continue
        tokens.append(token)
    if tokens:
        results.append(tokens)
    return results


def get_sentences(text):
    structure = []
    sentences = []
    sentence = u''
    place_holders = []
    tokens = diff.tokenize(text)
    position = 0
    for word_tokens in tokens_generator(tokens):
        if not word_tokens:
            structure.append([word_tokens, None])
            position += 1

        for tokens in split_token(word_tokens):
            word = u''.join(map(unicode, tokens))
            word = word.rstrip(SENTENCES_END)

            last_token = tokens[-1]
            sentence += word + last_token.trailing_whitespace

            place_holder = PlaceHolder(word, position)
            place_holders.append(place_holder)
            structure.append([tokens, place_holder])

            if last_token and last_token[-1] in SENTENCES_END:
                sentences.append(Sentence(sentence, place_holders))
                sentence = u''
                place_holders = []
            position += 1
    else:
        if sentence:
            sentences.append(Sentence(sentence, place_holders))
    return structure, sentences



#! coding: utf-8
from operator import itemgetter
from collections import defaultdict, OrderedDict
from pkg_resources import resource_filename

import nltk
import pymorphy2

from .analyzer import ExactAnalyzer, SubformsAnalyzer


morph = pymorphy2.MorphAnalyzer()
filename = resource_filename("ta4", "data/nltk/english.pickle")
tokenizer = nltk.data.load(filename)


def mark_with_words(words, text, analyzers={}):
    u"""
    Находит в тексте вхождения слов, при помощи анализаторов.
    Видоизменяет плейсхолдеры текста, добавляя маркеры фраз под которые подходит плейсхолдер.

    :param words: list text_analyze.sentence.Sentence
    :param text: list text_analyze.text.TextHtml
    """
    # Ключом будет являться, является ли задание - заданием на поиск точного вхождения
    analyzers = analyzers or {
        True: ExactAnalyzer(),
        False: SubformsAnalyzer(),
    }
    number = 1  # скозная нумерация маркеров
    for word in words:
        for sentence in text:
            analyzer = analyzers[word.is_exact_task]
            number = analyzer.mark(word, sentence, number)


def find_words(words, text):
    u"""
    Из размеченного текста(функцией mark_with_words), выбирает вхождения слов,
    и формирует список какая фраза сколько раз вошло в текст, согласно приоритетам:
     - точное вхождение, приоритетнее вхождения по словоформам("пластиковые окна" > "[купить] [пластиковое] [окно]")
     - вхождение с большим числом слов, приоритетнее ("купить пластивые окна" > "пластиковое окно")
     - при сравнении вхождений по словоформам при равенстве числа слов, приоритетнее то,
       в котором меньше звёздочек ("[купить] [пластиковое] [окно]" > "[купить] [*] [окно]")

    В результате получим подсчёт количества вхождений words в text, и дополненное задание
    """
    counter = {w.text: 0 for w in words}
    new_tasks = defaultdict(int)
    for sentence in text:
        markers = get_markers(sentence)
        # группируем маркеры по пересечению
        # в итоге получим кластера, которые хоть как то пересекаются
        for markers_chunk in group_markers(markers):
            if len(markers_chunk) == 1:
                # пересечений по маркерам нет
                marker_sentence, _ = markers_chunk[0]
                counter[marker_sentence.text] += 1
            else:
                # подсчитываем вхождения и добавляем попарное пересечение в задание
                length = len(markers_chunk) - 1
                for i, (marker_sentence, marker) in enumerate(markers_chunk):
                    counter[marker_sentence.text] += 1
                    if i < length:
                        _, next_marker = markers_chunk[i+1]
                        text = get_intersection(marker, next_marker, sentence)
                        new_tasks[text] += 1

    return counter, dict(new_tasks)


def get_markers(sentence):
    """
    Для предложения вернёт все маркеры, с диапазоноп позиций

    (u'купить пластиковые окна', {min: 2, max: 4})
    (u'пластиковые окна в москве', {min: 3, max: 6})
    """
    markers = OrderedDict()
    marker_id_to_sentence = {}
    for ph in sentence.place_holders:
        for marker in ph.markers:
            marker_id_to_sentence[marker.id] = marker.sentence
            if marker.id not in markers:
                markers[marker.id] = {'min': ph.position, 'max': ph.position}
            else:
                markers[marker.id]['max'] = ph.position
    result = [(marker_id_to_sentence[pk], positions) for pk, positions in markers.items()]
    return result


def group_markers(markers):
    result = []
    length = len(markers) - 1
    max_pos = 0
    for i, (_, marker) in enumerate(markers):
        result.append(markers[i])
        if max_pos < marker['max']:
            max_pos = marker['max']
        if i < length:
            _, next_marker = markers[i+1]
            # Проверяем, не началась ли другая группа маркеров
            if next_marker['min'] > max_pos:
                # сортируем по приоритету
                result = sorted(result, cmp=phrase_cmp, key=itemgetter(0), reverse=True)
                # отфильтровываем поглащённые
                result = merge_filter(result)
                if result:
                    yield result
                max_pos = next_marker['max']
                result = []
    result = sorted(result, cmp=phrase_cmp, key=itemgetter(0), reverse=True)
    result = merge_filter(result)
    if result:
        yield result


def merge_filter(markers):
    """
    Фильтрует маркеры на предмет полного поглощения, так как markers отсортированы по приоритету,
    функция выкинет маркеры, которые полностью поглощаются более приоритетными фразами
    """
    result = []
    for i, (sentence, marker) in enumerate(markers):
        # смотрим на предыдущие маркеры, входит ли в них i-ый
        for _, bigger_marker in result:
            if bigger_marker['min'] <= marker['min'] <= bigger_marker['max'] \
               and bigger_marker['min'] <= marker['max'] <= bigger_marker['max']:
                break
        else:
            result.append((sentence, marker))
    return result


def get_intersection(marker, next_marker, sentence):
    """
    Возвращает текст из пересечения маркеров
    """
    #  пересечение вида ------
    #                       ------
    if marker['min'] <= next_marker['min'] <= marker['max'] and next_marker['max'] > marker['max']:
        i, j = next_marker['min'], marker['max']
    #  пересечение вида ---
    #                 --------
    elif next_marker['min'] < marker['max'] and next_marker['max'] > marker['max']:
        i, j = marker['min'], marker['max']
    #  пересечение вида ---------
    #                     ----
    elif marker['min'] < next_marker['min'] < marker['max'] and marker['min'] < next_marker['max'] < marker['max']:
        i, j = next_marker['min'], next_marker['max']
    #  пересечение вида --------
    #               --------
    else:
        i, j = marker['min'], next_marker['max']
    phrase = []
    for ph in sentence.place_holders:
        if i <= ph.position <= j:
            phrase.append(ph.origin_word)

    return u' '.join(phrase)


def phrase_cmp(one, another):
    """
    Функция сравнения, для сортировки фраз по порядку их приоритета

    :param one: text_analyze.sentence.Sentence
    :param another: text_analyze.sentence.Sentence
    """
    # точное вхождение приоритетнее словоформ
    subform_cmp = cmp(one.is_exact_task, another.is_exact_task)
    if subform_cmp != 0:
        return subform_cmp
    # фраза с большим числом слов - приоритетнее
    len_cmp = cmp(len(one), len(another))
    if len_cmp != 0:
        return len_cmp

    # для словоформ с равным числом слов, приоритетное то, где меньше звёздочек
    if not one.is_exact_task and not another.is_exact_task:
        return cmp(one.count(u'[*]'), another.count(u'[*]')) * -1
    return len_cmp



#! coding: utf-8
from operator import attrgetter


class Sentence(object):
    def __init__(self, text, placeholders=None):
        self.text = text
        self.place_holders = placeholders or []
        if not self.place_holders:
            from .utils import get_sentences
            _, sentences = get_sentences(self.text)
            self.place_holders = sentences[0].place_holders

    def __len__(self):
        return len(self.place_holders)

    def __repr__(self):
        return self.text.encode('utf-8')

    @property
    def is_exact_task(self):
        """
        Если задание по словоформам - вернёт False(например "[пластиковые] [окна]")
        """
        return not any(map(attrgetter('is_subform_word'), self.place_holders))

    def count(self, word):
        return len([ph for ph in self.place_holders if ph.word == word])



# -*-coding: utf-8 -*-
import hashlib

from lexeme import Lexeme


class Marker(object):
    def __init__(self, sentence, position, marker_id):
        """
        marker_id - признак позволяющий найти
        """
        self.id = marker_id
        self.sentence = sentence
        self.position = position

    @property
    def hash(self):
        return hashlib.md5(self.sentence.text.encode('utf-8')).hexdigest()


class PlaceHolder(Lexeme):
    def __init__(self, word, position):
        super(PlaceHolder, self).__init__(word)
        self.position = position
        self.markers = []
        # этот плейсхолдер - словоформа
        self.is_subform_word = word.startswith('[') and word.endswith(']')

    def add_marker(self, position, word, marker_id):
        marker = Marker(word, position, marker_id)
        self.markers.append(marker)



#!/usr/bin/env python
# -*-coding: utf-8 -*-
import re
from operator import attrgetter

from . import morph


clean_word_regexp = re.compile(ur'^[^\w]*([-\w]+)[^\w]*$', re.U | re.I)


PARTS_OF_SPEECH = set([
    u'ADJF',  # имя прилагательное (полное)
    u'ADJS',  # имя прилагательное (краткое)
    u'ADVB',  # наречие
    u'COMP',  # компаратив
    u'CONJ',  # союз
    u'GRND',  # деепричастие
    u'INFN',  # глагол (инфинитив)
    u'INTJ',  # междометие
    u'NOUN',  # имя существительное
    u'NPRO',  # местоимение-существительное
    u'NUMR',  # числительное
    u'PRCL',  # частица
    u'PRED',  # предикатив
    u'PREP',  # предлог
    u'PRTF',  # причастие (полное)
    u'PRTS',  # причастие (краткое)
    u'VERB',  # глагол (личная форма)
])

IMPORTANT_PARTS_OF_SPEECH = PARTS_OF_SPEECH - set(['NPRO', 'PRED', 'PREP', 'CONJ', 'INTJ', 'PRCL'])


class GramInfo(object):
    __slots__ = ('normal_form', 'part_of_speech')

    def __init__(self, normal_form, part_of_speech):
        self.normal_form = normal_form
        self.part_of_speech = part_of_speech or u'NOUN'

    @property
    def is_important_pos(self):
        return self.part_of_speech in IMPORTANT_PARTS_OF_SPEECH


def get_gram_infos(word):
    results = morph.parse(word)
    return [GramInfo(r.normal_form, r.tag.POS) for r in results]


class Lexeme(object):
    __slots__ = ('word', 'origin_word', 'gram_infos')

    SPECIAL_WORD = '[*]'

    def __init__(self, word):
        clear_word = clean_word_regexp.sub(ur'\1', word)
        self.word = clear_word.upper()
        self.origin_word = clear_word.upper()
        self.gram_infos = get_gram_infos(clear_word)

    @property
    def is_special(self):
        return self.origin_word == self.SPECIAL_WORD

    @property
    def is_important(self):
        u"""Не важная часть речи, например предлоги или частицы, не учавствуют в сравнениях"""
        if self.is_special:
            return False
        return all(map(attrgetter('is_important_pos'), self.gram_infos))

    def get_all_normal_phorms(self):
        return set(map(attrgetter('normal_form'), self.gram_infos))

    def __repr__(self):
        return self.word.encode('utf-8')
